import type { Source, Target, Node, Bound, KeyValueStore } from "@canvas-js/okra";
import { EventEmitter } from "@libp2p/interface/events";
import { Logger } from "@libp2p/logger";
import { Schema } from "@ipld/schema/schema-schema";
import type { Message, MessageSigner, Awaitable } from "@canvas-js/interfaces";
import { Signature } from "@canvas-js/signed-cid";
import { Mempool } from "./Mempool.js";
export interface ReadOnlyTransaction {
    messages: Omit<KeyValueStore, "set" | "delete"> & Source;
    parents: Omit<KeyValueStore, "set" | "delete">;
    ancestors?: Omit<KeyValueStore, "set" | "delete">;
}
export interface ReadWriteTransaction {
    messages: KeyValueStore & Target;
    parents: KeyValueStore;
    ancestors?: KeyValueStore;
}
export type GossipLogConsumer<Payload = unknown, Result = void> = (id: string, signature: Signature, message: Message<Payload>) => Awaitable<Result>;
export interface GossipLogInit<Payload = unknown, Result = void> {
    topic: string;
    apply: GossipLogConsumer<Payload, Result>;
    validate: ((payload: unknown) => payload is Payload) | {
        schema: string | Schema;
        name: string;
    };
    signer?: MessageSigner<Payload>;
    indexAncestors?: boolean;
}
export type GossipLogEvents<Payload = unknown, Result = void> = {
    message: CustomEvent<{
        id: string;
        signature: Signature;
        message: Message<Payload>;
        result: Result;
    }>;
    commit: CustomEvent<{
        root: Node;
    }>;
    sync: CustomEvent<{
        peer: string | null;
    }>;
};
export declare abstract class AbstractGossipLog<Payload = unknown, Result = unknown> extends EventEmitter<GossipLogEvents<Payload, Result>> {
    #private;
    private static bound;
    abstract close(): Promise<void>;
    protected abstract entries(lowerBound?: Bound<Uint8Array> | null, upperBound?: Bound<Uint8Array> | null, options?: {
        reverse?: boolean;
    }): AsyncIterable<[key: Uint8Array, value: Uint8Array]>;
    protected abstract read<T>(callback: (txn: ReadOnlyTransaction) => Awaitable<T>, options?: {
        targetId?: string;
    }): Promise<T>;
    protected abstract write<T>(callback: (txn: ReadWriteTransaction) => Awaitable<T>, options?: {
        sourceId?: string;
    }): Promise<T>;
    readonly topic: string;
    readonly indexAncestors: boolean;
    readonly signer: MessageSigner<Payload>;
    protected readonly log: Logger;
    protected readonly mempool: Mempool<{
        signature: Signature;
        message: Message<Payload>;
    }>;
    protected constructor(init: GossipLogInit<Payload, Result>);
    replay(): Promise<void>;
    iterate(lowerBound?: {
        id: string;
        inclusive: boolean;
    } | null, upperBound?: {
        id: string;
        inclusive: boolean;
    } | null, options?: {
        reverse?: boolean;
    }): AsyncIterable<[id: string, signature: Signature, message: Message<Payload>]>;
    encode(signature: Signature, message: Message<Payload>): [key: Uint8Array, value: Uint8Array];
    decode(value: Uint8Array): [id: string, signature: Signature, message: Message<Payload>];
    getClock(): Promise<[clock: number, parents: string[]]>;
    has(id: string): Promise<boolean>;
    get(id: string): Promise<[signature: Signature, message: Message<Payload>] | [null, null]>;
    private getParents;
    append(payload: Payload, options?: {
        signer?: MessageSigner<Payload>;
    }): Promise<{
        id: string;
        signature: Signature;
        message: Message<Payload>;
        result: Result;
    }>;
    insert(signature: Signature, message: Message<Payload>): Promise<{
        id: string;
    }>;
    getAncestors(id: string, atOrBefore: number): Promise<string[]>;
    isAncestor(id: string, ancestor: string): Promise<boolean>;
    static isAncestor(txn: ReadOnlyTransaction, id: string, ancestor: string, visited?: Set<string>): Promise<boolean>;
    sync(source: Source, options?: {
        sourceId?: string;
    }): Promise<{
        root: Node;
    }>;
    serve(callback: (source: Source) => Promise<void>, options?: {
        targetId?: string;
    }): Promise<void>;
}
