var __classPrivateFieldSet = (this && this.__classPrivateFieldSet) || function (receiver, state, value, kind, f) {
    if (kind === "m") throw new TypeError("Private method is not writable");
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a setter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot write private member to an object whose class did not declare it");
    return (kind === "a" ? f.call(receiver, value) : f ? f.value = value : state.set(receiver, value)), value;
};
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var _AbstractGossipLog_instances, _AbstractGossipLog_codec, _AbstractGossipLog_apply, _AbstractGossipLog_getAncestors, _AbstractGossipLog_insert;
import { CustomEvent, EventEmitter } from "@libp2p/interface/events";
import { logger } from "@libp2p/logger";
import { equals } from "uint8arrays";
import { bytesToHex as hex } from "@noble/hashes/utils";
import * as cbor from "@ipld/dag-cbor";
import { fromDSL } from "@ipld/schema/from-dsl.js";
import { create } from "@ipld/schema/typed.js";
import { verifySignature } from "@canvas-js/signed-cid";
import { Ed25519Signer } from "./Ed25519Signer.js";
import { Mempool } from "./Mempool.js";
import { Driver } from "./sync/driver.js";
import { decodeClock } from "./clock.js";
import { decodeId, encodeId, decodeSignedMessage, encodeSignedMessage, getNextClock, KEY_LENGTH } from "./schema.js";
import { assert, topicPattern, cborNull, getAncestorClocks } from "./utils.js";
export class AbstractGossipLog extends EventEmitter {
    constructor(init) {
        super();
        _AbstractGossipLog_instances.add(this);
        this.mempool = new Mempool();
        _AbstractGossipLog_codec.set(this, void 0);
        _AbstractGossipLog_apply.set(this, void 0);
        assert(topicPattern.test(init.topic), "invalid topic (must match [a-zA-Z0-9\\.\\-])");
        this.topic = init.topic;
        __classPrivateFieldSet(this, _AbstractGossipLog_apply, init.apply, "f");
        if (typeof init.validate === "function") {
            const { validate } = init;
            __classPrivateFieldSet(this, _AbstractGossipLog_codec, {
                toTyped: (obj) => (validate(obj) ? obj : undefined),
                toRepresentation: (obj) => (validate(obj) ? obj : undefined),
            }, "f");
        }
        else {
            const { schema, name } = init.validate;
            __classPrivateFieldSet(this, _AbstractGossipLog_codec, create(typeof schema === "string" ? fromDSL(schema) : schema, name), "f");
        }
        this.indexAncestors = init.indexAncestors ?? false;
        this.signer = init.signer ?? new Ed25519Signer();
        this.log = logger(`canvas:gossiplog:[${this.topic}]`);
    }
    async replay() {
        await this.read(async (txn) => {
            for await (const [key, value] of txn.messages.entries()) {
                const [id, signature, message] = this.decode(value);
                assert(id === decodeId(key), "expected id === decodeId(key)");
                await __classPrivateFieldGet(this, _AbstractGossipLog_apply, "f").apply(txn, [id, signature, message]);
            }
        });
    }
    async *iterate(lowerBound = null, upperBound = null, options = {}) {
        for await (const [key, value] of this.entries(AbstractGossipLog.bound(lowerBound?.id, lowerBound?.inclusive), AbstractGossipLog.bound(upperBound?.id, upperBound?.inclusive), options)) {
            const [id, signature, message] = this.decode(value);
            assert(id === decodeId(key), "expected id === decodeId(key)");
            yield [id, signature, message];
        }
    }
    encode(signature, message) {
        assert(message.topic === this.topic, "invalid message topic");
        return encodeSignedMessage(signature, message, __classPrivateFieldGet(this, _AbstractGossipLog_codec, "f"));
    }
    decode(value) {
        const [id, signature, message] = decodeSignedMessage(value, __classPrivateFieldGet(this, _AbstractGossipLog_codec, "f"));
        assert(message.topic === this.topic, "invalid message topic");
        return [id, signature, message];
    }
    async getClock() {
        const parents = await this.read((txn) => this.getParents(txn));
        const clock = getNextClock(parents);
        return [clock, parents.map(decodeId)];
    }
    async has(id) {
        return await this.read(({ messages }) => messages.get(encodeId(id)) !== null);
    }
    async get(id) {
        const value = await this.read(({ messages }) => messages.get(encodeId(id)));
        if (value === null) {
            return [null, null];
        }
        const [_, signature, message] = this.decode(value);
        return [signature, message];
    }
    async getParents(txn) {
        const parents = [];
        for await (const [key, value] of txn.parents.entries()) {
            assert(key.byteLength === KEY_LENGTH, "internal error (expected key.byteLength === KEY_LENGTH)");
            assert(equals(value, cborNull), "internal error (unexpected parent entry value)");
            parents.push(key);
        }
        return parents;
    }
    async append(payload, options = {}) {
        const signer = options.signer ?? this.signer;
        const { id, signature, message, result, root } = await this.write(async (txn) => {
            const parents = await this.getParents(txn);
            const clock = getNextClock(parents);
            const message = { topic: this.topic, clock, parents: parents.map(decodeId), payload };
            const signature = await signer.sign(message);
            const [key, value] = this.encode(signature, message);
            const id = decodeId(key);
            this.log("appending message %s: %O", id, message);
            const result = await __classPrivateFieldGet(this, _AbstractGossipLog_instances, "m", _AbstractGossipLog_insert).call(this, txn, id, signature, message, [key, value]);
            const root = await txn.messages.getRoot();
            return { id, signature, message, result, root };
        });
        this.dispatchEvent(new CustomEvent("commit", { detail: { topic: this.topic, root } }));
        this.log("commited root %s", hex(root.hash));
        return { id, signature, message, result };
    }
    async insert(signature, message) {
        verifySignature(signature, message);
        const { id, root } = await this.write(async (txn) => {
            const [key, value] = this.encode(signature, message);
            const id = decodeId(key);
            this.log("inserting message %s", id);
            const missingParents = new Set();
            this.log("looking up %s parents", message.parents.length);
            for (const parentId of message.parents) {
                const parent = await txn.messages.get(encodeId(parentId));
                if (parent === null) {
                    this.log("missing parent %s", parentId);
                    missingParents.add(parentId);
                }
                else {
                    this.log("found parent %s", parentId);
                }
            }
            if (missingParents.size > 0) {
                this.log("missing %d/%d parents", missingParents.size, message.parents.length);
                this.mempool.add(id, { signature, message }, missingParents);
                return { id };
            }
            await __classPrivateFieldGet(this, _AbstractGossipLog_instances, "m", _AbstractGossipLog_insert).call(this, txn, id, signature, message, [key, value]);
            const root = await txn.messages.getRoot();
            return { id, root };
        });
        if (root !== undefined) {
            this.dispatchEvent(new CustomEvent("commit", { detail: { topic: this.topic, root } }));
            this.log("commited root %s", hex(root.hash));
        }
        return { id };
    }
    async getAncestors(id, atOrBefore) {
        const results = new Set();
        await this.read((txn) => __classPrivateFieldGet(this, _AbstractGossipLog_instances, "m", _AbstractGossipLog_getAncestors).call(this, txn, encodeId(id), atOrBefore, results));
        this.log("getAncestors of %s atOrBefore %d: %o", id, atOrBefore, results);
        return Array.from(results).sort();
    }
    async isAncestor(id, ancestor) {
        return await this.read((txn) => AbstractGossipLog.isAncestor(txn, id, ancestor));
    }
    static async isAncestor(txn, id, ancestor, visited = new Set()) {
        assert(txn.ancestors !== undefined, "expected txn.ancestors !== undefined");
        if (id === ancestor) {
            return true;
        }
        const ancestorKey = encodeId(ancestor);
        const [ancestorClock] = decodeClock(ancestorKey);
        const key = encodeId(id);
        const [clock] = decodeClock(key);
        if (clock <= ancestorClock) {
            return false;
        }
        const index = Math.floor(Math.log2(Number(clock - ancestorClock)));
        const value = await txn.ancestors.get(key);
        assert(value !== null, "key not found in ancestor index");
        const links = cbor.decode(value);
        for (const key of links[index]) {
            const id = decodeId(key);
            if (visited.has(id)) {
                continue;
            }
            visited.add(id);
            const isAncestor = await AbstractGossipLog.isAncestor(txn, id, ancestor, visited);
            if (isAncestor) {
                return true;
            }
        }
        return false;
    }
    async sync(source, options = {}) {
        const root = await this.write(async (txn) => {
            const driver = new Driver(this.topic, source, txn.messages);
            for await (const [key, value] of driver.sync()) {
                const [id, signature, message] = this.decode(value);
                assert(id === decodeId(key), "expected id === decodeId(key)");
                verifySignature(signature, message);
                const existingMessage = await txn.messages.get(key);
                if (existingMessage === null) {
                    for (const parent of message.parents) {
                        const parentKey = encodeId(parent);
                        const parentValue = await txn.messages.get(parentKey);
                        if (parentValue === null) {
                            this.log.error("missing parent %s of message %s: %O", parent, id, message);
                        }
                    }
                    await __classPrivateFieldGet(this, _AbstractGossipLog_instances, "m", _AbstractGossipLog_insert).call(this, txn, id, signature, message, [key, value]);
                }
            }
            return await txn.messages.getRoot();
        }, options);
        this.dispatchEvent(new CustomEvent("sync", { detail: { peer: options.sourceId ?? null } }));
        this.dispatchEvent(new CustomEvent("commit", { detail: { root } }));
        this.log("commited root %s", hex(root.hash));
        return { root };
    }
    async serve(callback, options = {}) {
        await this.read((txn) => callback(txn.messages), options);
    }
}
_AbstractGossipLog_codec = new WeakMap(), _AbstractGossipLog_apply = new WeakMap(), _AbstractGossipLog_instances = new WeakSet(), _AbstractGossipLog_getAncestors = async function _AbstractGossipLog_getAncestors(txn, key, atOrBefore, results, visited = new Set()) {
    assert(txn.ancestors !== undefined, "expected txn.ancestors !== undefined");
    assert(atOrBefore > 0, "expected atOrBefore > 0");
    const [clock] = decodeClock(key);
    assert(atOrBefore < Number(clock), "expected atOrBefore < clock");
    const index = Math.floor(Math.log2(Number(clock) - atOrBefore));
    const value = await txn.ancestors.get(key);
    if (value === null) {
        throw new Error(`key ${decodeId(key)} not found in ancestor index`);
    }
    const links = cbor.decode(value);
    for (const ancestorKey of links[index]) {
        const [ancestorClock] = decodeClock(ancestorKey);
        const ancestorId = decodeId(ancestorKey);
        if (Number(ancestorClock) <= atOrBefore) {
            results.add(ancestorId);
        }
        else if (visited.has(ancestorId)) {
            return;
        }
        else {
            visited.add(ancestorId);
            await __classPrivateFieldGet(this, _AbstractGossipLog_instances, "m", _AbstractGossipLog_getAncestors).call(this, txn, ancestorKey, atOrBefore, results, visited);
        }
    }
}, _AbstractGossipLog_insert = async function _AbstractGossipLog_insert(txn, id, signature, message, [key, value] = this.encode(signature, message)) {
    this.log("applying %s %O", id, message);
    const result = await __classPrivateFieldGet(this, _AbstractGossipLog_apply, "f").apply(txn, [id, signature, message]);
    this.dispatchEvent(new CustomEvent("message", { detail: { id, signature, message, result } }));
    await txn.messages.set(key, value);
    await txn.parents.set(key, cborNull);
    for (const parentId of message.parents) {
        await txn.parents.delete(encodeId(parentId));
    }
    if (this.indexAncestors) {
        assert(txn.ancestors !== undefined, "expected txn.ancestors !== undefined");
        const ancestorClocks = Array.from(getAncestorClocks(message.clock));
        const ancestorLinks = new Array(ancestorClocks.length);
        for (const [i, ancestorClock] of ancestorClocks.entries()) {
            if (i === 0) {
                ancestorLinks[i] = message.parents.map(encodeId);
            }
            else {
                const links = new Set();
                for (const child of ancestorLinks[i - 1]) {
                    const [childClock] = decodeClock(child);
                    if (Number(childClock) <= ancestorClock) {
                        links.add(decodeId(child));
                    }
                    else {
                        assert(Number(childClock) <= ancestorClocks[i - 1], "expected childClock <= ancestorClocks[i - 1]");
                        await __classPrivateFieldGet(this, _AbstractGossipLog_instances, "m", _AbstractGossipLog_getAncestors).call(this, txn, child, ancestorClock, links);
                    }
                }
                ancestorLinks[i] = Array.from(links).map(encodeId);
            }
        }
        await txn.ancestors.set(key, cbor.encode(ancestorLinks));
    }
    for (const [childId, signedMessage] of this.mempool.observe(id)) {
        await __classPrivateFieldGet(this, _AbstractGossipLog_instances, "m", _AbstractGossipLog_insert).call(this, txn, childId, signedMessage.signature, signedMessage.message);
    }
    return result;
};
AbstractGossipLog.bound = (id = null, inclusive = true) => id === null ? null : { key: encodeId(id), inclusive };
