import { base32hex } from "multiformats/bases/base32";
import { sha256 } from "@noble/hashes/sha256";
import * as cbor from "@ipld/dag-cbor";
import { fromDSL } from "@ipld/schema/from-dsl.js";
import { create } from "@ipld/schema/typed.js";
import { lessThan } from "@canvas-js/okra";
import { decodeClock, encodeClock } from "./clock.js";
import { assert } from "./utils.js";
const schema = fromDSL(`
type SignedMessage struct {
	signature Signature
	topic String
	parents [Bytes]
	payload any
} representation tuple

type Signature struct {
	type String
	publicKey Bytes
	signature Bytes
	cid &Message
} representation tuple
`);
const { toTyped: toSignedMessage, toRepresentation: fromSignedMessage } = create(schema, "SignedMessage");
export function decodeSignedMessage(value, { toTyped }) {
    const signedMessage = toSignedMessage(cbor.decode(value));
    assert(signedMessage !== undefined, "error decoding message (internal error)");
    const { signature, topic } = signedMessage;
    const clock = getNextClock(signedMessage.parents);
    const parents = signedMessage.parents ?? [];
    assert(parents.every((id, i) => i === 0 || lessThan(parents[i - 1], id)), "unsorted parents array");
    const payload = toTyped(signedMessage.payload);
    assert(payload !== undefined, "error decoding message (invalid payload)");
    const id = decodeId(getKey(clock, sha256(value)));
    return [id, signature, { topic, clock, parents: parents.map(decodeId), payload }];
}
export function encodeSignedMessage(signature, message, { toRepresentation }) {
    const parents = message.parents.sort().map(encodeId);
    assert(getNextClock(parents) === message.clock, "error encoding message (invalid clock)");
    const payload = toRepresentation(message.payload);
    assert(payload !== undefined, "error encoding message (invalid payload)");
    const signedMessage = {
        signature,
        topic: message.topic,
        parents: parents,
        payload: payload,
    };
    const signedMessageRepresentation = fromSignedMessage(signedMessage);
    assert(signedMessageRepresentation !== undefined, "error encoding message (internal error)");
    const value = cbor.encode(signedMessageRepresentation);
    const key = getKey(message.clock, sha256(value));
    return [key, value];
}
export function getNextClock(parents) {
    let max = 0;
    for (const key of parents) {
        assert(key.byteLength === KEY_LENGTH, "expected key.byteLength === KEY_LENGTH");
        const [clock] = decodeClock(key);
        if (clock > max) {
            max = clock;
        }
    }
    return max + 1;
}
export const KEY_LENGTH = 20;
export const MIN_MESSAGE_ID = "00000000000000000000000000000000";
export const MAX_MESSAGE_ID = "vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv";
function getKey(clock, hash) {
    const key = new Uint8Array(KEY_LENGTH);
    const encodingLength = encodeClock(key, clock);
    key.set(hash.subarray(0, KEY_LENGTH - encodingLength), encodingLength);
    return key;
}
export const encodeId = (id) => base32hex.baseDecode(id);
export const decodeId = (key) => base32hex.baseEncode(key);
